/**
 * Author: Andrey Nechypurenko (andreynech@gmail.com)
 */

#include "CapturerThread.h"
#include "VideoFrame.h"
#include "Decoder.h"
#include "yuv2rgb.h"
#include <QMutexLocker>
#include <QString>

#define MPEG_BUF_SIZE (1024*128)


CapturerThread::CapturerThread(VideoFrame *vf) 
  : should_stop(true), 
    videoframe(vf),
    boardnum(0),
    encode(MIDSRAY_ENCODE_MPEG4),
    size(0),
    vidsys(MIDSRAY_VIDSYS_PAL),
    bitrate(150000), // 150Kbit/s
    iframesonly(0)
{
  n_buffers = 1;
  this->buffers = (buffer*)calloc(1, sizeof(*buffers));

  if(!(this->buffers))
    throw(std::bad_alloc());

  this->buffers[0].length = MPEG_BUF_SIZE;
  this->buffers[0].start = malloc(MPEG_BUF_SIZE);

  if(!buffers[0].start)
    throw(std::bad_alloc());
}

CapturerThread::~CapturerThread()
{
  if(this->isRunning())
    {
      this->shouldStop();
      sleep(1);
    }
}


void 
CapturerThread::shouldStop()
{
  this->should_stop = true;
}


int
CapturerThread::getFrameWidth()
{
#ifndef FILESOURCE
  int w;
  int res = SN_GetWidth(this->boardnum, &w);
  if(MIDSRAY_SUCCESS != res)
    {
      QString errmsg;
      this->errno_string(res, "getFrameWidth", errmsg);
      emit error(errmsg);
    }

  return w;
#else
  return 720; // PAL
#endif
}


int
CapturerThread::getFrameHeight()
{
#ifndef FILESOURCE
  int h;
  int res = SN_GetHeight(this->boardnum, &h);
  if(MIDSRAY_SUCCESS != res)
    {
      QString errmsg;
      this->errno_string(res, "getFrameHeight", errmsg);
      emit error(errmsg);
    }

  return h;
#else
  return 576; // PAL
#endif
}

void
CapturerThread::errno_string(int errorcode,
                             const QString &context,
                             QString &strerr) const
{
  strerr = context;

  switch(errorcode)
    {
    case MIDSRAY_ERR_CAPTURING:
      strerr += tr(" already capturing/recording");
      break;
    case MIDSRAY_ERR_NOTCAPTURING:
      strerr += tr(" not capturing/recording");
      break;
    case MIDSRAY_ERR_RUNNING:
      strerr += tr(" attempt to set configuration while graph running");
      break;
    case MIDSRAY_ERR_SNAPSHOT:
      strerr += tr(" currently displaying snapshot");
      break;
    case MIDSRAY_ERR_INVALIDFILE:
      strerr += tr(" invalid file");
      break;
    case MIDSRAY_ERR_DIRECTX:
      strerr += tr(" directX error (other specific directX errors are possible)");
      break;
    case MIDSRAY_ERR_NODEVICE:
      strerr += tr(" device not present");
      break;
    case MIDSRAY_ERR_NOTINITIALIZED:
      strerr += tr(" initialization function failed or not called");
      break;
    case MIDSRAY_ERR_INVALIDMODE:
      strerr += tr(" invalid mode(eg. record in MJPEG CAPTURE mode)");
      break;
    case MIDSRAY_ERR_NOMEM:
      strerr += tr(" no memory");
      break;
    case MIDSRAY_ERR_INVALIDPARAM:
      strerr += tr(" invalid parameter");
      break;
    case MIDSRAY_ERR_INVALIDUSERDATA:
      strerr += tr(" user data contains MPEG start code");
      break;
    case MIDSRAY_ERR_NONEWDATA:
      strerr += tr(" no new data");
      break;
    case MIDSRAY_ERR_UNSPECIFIED:
    default:
      strerr += tr(" unspecified error");
      break;
    }
}


int
CapturerThread::readFrame()
{
  int res;
  frame_info_t fi;;

  res = SN_GetOneFrame(this->boardnum,
                       (UINT8*)this->buffers[0].start,
                       MPEG_BUF_SIZE,
                       &fi);

  if(MIDSRAY_SUCCESS != res)
    {
      // device not ready, wait and try again later
      //printf("rt %d  %d %d\n", rt, fi.data_length,0);//fi.FrameType);
      return 0;
    }
  //printf("got this much data %d\n", fi.data_length);
  this->buffers[0].length = fi.data_length;

  // Return number of bytes transferred.
  return fi.data_length;
}


void 
CapturerThread::run()
{
  QString errmsg;
  int res;

  res = SN_OpenBoard(this->boardnum);
  if(MIDSRAY_SUCCESS != res)
    {
      QString errmsg;
      this->errno_string(res, tr("board open failed"), errmsg);
      emit error(errmsg);
      return;
    }
  // set video source: 0=S-Video, 1=Composite
  SN_SetVideoSource(this->boardnum, 1);
  SN_SetBasicChipSettings(this->boardnum,
                          this->encode,
                          this->size,
                          this->vidsys,
                          this->bitrate,
                          this->iframesonly);
  res = SN_StartStream(this->boardnum);
  if(MIDSRAY_SUCCESS != res)
    {
      errno_string(res, tr("startCapturing"), errmsg);
      emit error(errmsg);
      return;
    }
  this->videoframe->width = this->getFrameWidth();
  this->videoframe->height = this->getFrameHeight();
  // calculate texture size which should be the power of 2
  for(this->videoframe->texwidth = 1;
      this->videoframe->texwidth < this->videoframe->width;
      this->videoframe->texwidth *= 2);
  for(this->videoframe->texheight = 1;
      this->videoframe->texheight < this->videoframe->height;
      this->videoframe->texheight *= 2);

  this->videoframe->texdata =
    new unsigned char[this->videoframe->texwidth*this->videoframe->texheight*4];  memset(this->videoframe->texdata,
         0x80808080,
         this->videoframe->texwidth * this->videoframe->texheight * 4);

  emit setupTex();

  Decoder decoder(this->getFrameWidth(), this->getFrameHeight(), "mpeg4");

  unsigned char *picture;
  uint8_t *buff_ptr;
  int got_picture;
  ssize_t size;
  int len;

  this->should_stop = false;
  while(!(this->should_stop))
    {
      if(0 == this->readFrame())
        {
          usleep(10000);
        }
      else
        {
          buff_ptr = (uint8_t*)(this->buffers[0].start);
          size = this->buffers[0].length;
          while(size > 0)
            {
              len = decoder.decode(buff_ptr, size, &picture, &got_picture);
              if(len == 0)
                break;

              if(got_picture)
                {
                  QMutexLocker lock(&(this->videoframe->mutex));
                  YUV420_to_RGB32(this->videoframe->width,
                                  this->videoframe->height,
                                  (char*)picture,
                                  (char*)(this->videoframe->texdata));
                  emit newFrame();
                }
              size -= len;
              buff_ptr += len;
            }
        }
    }

  SN_StopStream(this->boardnum);
  SN_CloseBoard(this->boardnum);
}
