/**
 * Author: Andrey Nechypurenko (andreynech@gmail.com)
 */

#include "Decoder.h"
#include <iostream>

const char* 
xvid_error_to_string(int retcode)
{
  switch(retcode)
    {
    case XVID_ERR_FAIL: return "general fault";
    case XVID_ERR_MEMORY: return "memory allocation error";
    case XVID_ERR_FORMAT: return "file format error";
    case XVID_ERR_VERSION: return "structure version not supported";
    case XVID_ERR_END: return "encoder only; end of stream reached";
    }
  return "unknown error code returned";
}


Decoder::Decoder(int frame_w, int frame_h, const char *codec_name) :
  width(frame_w), height(frame_h), dec_handle(NULL)
{
  int status = this->dec_init();
  if(status)
    {
      std::cout << "Xvid error: " << xvid_error_to_string(status) << std::endl;
    }
  this->out_buffer = (unsigned char*)new char[this->width * this->height * 4];
}


Decoder::~Decoder()
{
  xvid_decore(this->dec_handle, XVID_DEC_DESTROY, NULL, NULL);
  if(this->out_buffer)
    delete this->out_buffer;
}

int
Decoder::dec_init(int use_assembler)
{
  xvid_gbl_init_t   xvid_gbl_init;
  xvid_dec_create_t xvid_dec_create;

  /* Reset the structure with zeros */
  memset(&xvid_gbl_init, 0, sizeof(xvid_gbl_init_t));
  memset(&xvid_dec_create, 0, sizeof(xvid_dec_create_t));

  /*------------------------------------------------------------------------
   * XviD core initialization
   *----------------------------------------------------------------------*/

  /* Version */
  xvid_gbl_init.version = XVID_VERSION;

  /* Assembly setting */
  if(use_assembler)
#ifdef ARCH_IS_IA64
    xvid_gbl_init.cpu_flags = XVID_CPU_FORCE | XVID_CPU_IA64;
#else
  xvid_gbl_init.cpu_flags = 0;
#endif
  else
    xvid_gbl_init.cpu_flags = XVID_CPU_FORCE;

  xvid_gbl_init.debug = 0;

  int ret = xvid_global(NULL, 0, &xvid_gbl_init, NULL);
  if(ret < 0)
    return ret;

  /*------------------------------------------------------------------------
   * XviD encoder initialization
   *----------------------------------------------------------------------*/

  /* Version */
  xvid_dec_create.version = XVID_VERSION;

  /*
   * Image dimensions -- set to 0, xvidcore will resize when ever it is
   * needed
   */
  xvid_dec_create.width = 0;
  xvid_dec_create.height = 0;

  ret = xvid_decore(NULL, XVID_DEC_CREATE, &xvid_dec_create, NULL);
  if(ret < 0)
    return ret;

  this->dec_handle = xvid_dec_create.handle;

  return ret;
}


int 
Decoder::decode(unsigned char *frames,
                size_t size,
                unsigned char **picture, 
                int *got_picture)
{
  xvid_dec_frame_t xvid_dec_frame;
  xvid_dec_stats_t xvid_dec_stats;

  /* Reset all structures */
  memset(&xvid_dec_frame, 0, sizeof(xvid_dec_frame_t));
  memset(&xvid_dec_stats, 0, sizeof(xvid_dec_stats_t));

  /* Set version */
  xvid_dec_frame.version = XVID_VERSION;
  xvid_dec_stats.version = XVID_VERSION;

  /* No general flags to set */
  xvid_dec_frame.general          = 0;

  /* Input stream */
  xvid_dec_frame.bitstream        = frames;
  xvid_dec_frame.length           = int(size);

  /* Output frame structure */
  xvid_dec_frame.output.plane[0]  = this->out_buffer;
  xvid_dec_frame.output.stride[0] = this->width;
  xvid_dec_frame.output.csp = XVID_CSP_I420;

  int ret = xvid_decore(this->dec_handle, 
                        XVID_DEC_DECODE, 
                        &xvid_dec_frame, 
                        &xvid_dec_stats);

  if(xvid_dec_stats.type == XVID_TYPE_VOL)
    {
      *got_picture = 0;
      if(this->width < xvid_dec_stats.data.vol.width ||
         this->height < xvid_dec_stats.data.vol.height)
        {
          this->width = xvid_dec_stats.data.vol.width;
          this->height = xvid_dec_stats.data.vol.height;
          if(this->out_buffer)
            delete this->out_buffer;
          this->out_buffer = 
            (unsigned char*)new char[this->width * this->height * 4];
        }
    }
  else
    *got_picture = 1;

  *picture = this->out_buffer;

  return ret;
}
