/**
 * Author: Andrey Nechypurenko (andreynech@gmail.com)
 */

#include <math.h>
#include <QApplication>
#include <QMouseEvent>
#include "GLWidget.h"


GLWidget::GLWidget(QWidget *parent)
  : QGLWidget(parent)
{
  object = 0;
  xRot = 0;
  yRot = 0;
  zRot = 0;

  this->texture[0] = 65535;
}


GLWidget::~GLWidget()
{
  makeCurrent();
  glDeleteLists(object, 1);
}


QSize GLWidget::minimumSizeHint() const
{
  return QSize(50, 50);
}


QSize GLWidget::sizeHint() const
{
  return QSize(400, 400);
}


void GLWidget::setXRotation(int angle)
{
  normalizeAngle(&angle);
  if (angle != xRot) {
    xRot = angle;
    emit xRotationChanged(angle);
    updateGL();
  }
}


void GLWidget::setYRotation(int angle)
{
  normalizeAngle(&angle);
  if (angle != yRot) {
    yRot = angle;
    emit yRotationChanged(angle);
    updateGL();
  }
}


void GLWidget::setZRotation(int angle)
{
  normalizeAngle(&angle);
  if (angle != zRot) {
    zRot = angle;
    emit zRotationChanged(angle);
    updateGL();
  }
}


void GLWidget::initializeGL()
{
  qglClearColor(QApplication::palette().color(QPalette::Active, 
                                              QPalette::Window));
  glShadeModel(GL_FLAT);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_CULL_FACE);
  glEnable(GL_TEXTURE_2D); // Enable Texture Mapping
}


void 
GLWidget::setupTex(unsigned int framew,
                   unsigned int frameh,
                   unsigned int texw,
                   unsigned int texh,
                   unsigned char *frame)
{
  this->frameWidth = framew;
  this->frameHeight = frameh;
  this->bufferWidth = texw;
  this->bufferHeight = texh;
  if(this->texture[0] == 65535)
    {// create texture
      glGenTextures(1, &texture[0]);
      // 2d texture (x and y size)
      glBindTexture(GL_TEXTURE_2D, texture[0]);
      // scale linearly when image bigger than texture
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
      // scale linearly when image smalled than texture
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
      // 2d texture, level of detail 0 (normal), 3 components (red,
      // green, blue), x size from image, y size from image,
      // border 0 (normal), rgb color data, unsigned byte data, and
      // finally the data itself.
      glTexImage2D(GL_TEXTURE_2D,    // target
                   0,                // level
                   GL_RGBA,           // internalformat
                   this->bufferWidth,       // width
                   this->bufferHeight,      // height
                   0,                // border
                   GL_RGBA,          // format
                   GL_UNSIGNED_BYTE, // type
                   frame             // pixels
                   );
      int err = glGetError();
      switch(err)
        {
        case GL_NO_ERROR:
          //qDebug("Video texture created successful");
          break;
        case GL_INVALID_ENUM:
          qDebug("GL_INVALID_ENUM error creating texture: target is not GL_TEXTURE_2D or GL_PROXY_TEXTURE_2D; or");
          qDebug("format is not an accepted format constant. Format constants other than STENCIL_INDEX and GL_DEPTH_COMPONENT are accepted; or");
          qDebug("type is not a type constant; or");
          qDebug("type is GL_BITMAP and format is not GL_COLOR_INDEX.");
          break;
        case GL_INVALID_VALUE:
          {
          qDebug("GL_INVALID_VALUE error creating texture: level is less than 0; or");
          qDebug("level is greater than log max, where max is the returned value of GL_MAX_TEXTURE_SIZE; or");
          qDebug("internal format is not 1, 2, 3, 4, or one of the accepted resolution and format symbolic constants; or");
          qDebug("width or height is less than 0 or greater than 2 + GL_MAX_TEXTURE_SIZE, or if either cannot be represented as 2k+2(border) for some integer value of k; or");
          qDebug("border is not 0 or 1.");
          qDebug("GL_MAX_TEXTURE_SIZE = %i", GL_MAX_TEXTURE_SIZE);
          GLint texSize; 
          glGetIntegerv(GL_MAX_TEXTURE_SIZE, &texSize);
          qDebug("glGetIntegerv(GL_MAX_TEXTURE_SIZE, &texSize) returned %i", 
                 texSize);
          }
          break;
        case GL_INVALID_OPERATION:
          qDebug("GL_INVALID_OPERATION error creating texture: glTexImage2D is executed between the execution of glBegin and the corresponding execution of glEnd.");
          break;
        default:
          qDebug("glTexImage2D returned %i", err);
        }
      object = makeObject();
    }
}


void 
GLWidget::newFrame(unsigned char *frame)
{
  /*
  static bool save = true;
  if(save)
    {
      FILE *f = fopen("tex.rgb", "wb");
      fwrite(frame, 
             this->bufferWidth*this->bufferHeight*3, 
             1, f);
      fclose(f);
      save = false;
    }
  */
  // modify texture
  glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 
                  this->frameWidth, this->frameHeight, 
                  //this->bufferWidth, this->bufferHeight, 
                  GL_RGBA, GL_UNSIGNED_BYTE, 
                  frame);
  updateGL();
}


void 
GLWidget::paintGL()
{
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glLoadIdentity();
  glTranslatef(0.0f, 0.0f, -2.5f);
  //glTranslated(0.0, 0.0, -10.0);
  glRotated(xRot / 16.0, 1.0, 0.0, 0.0);
  glRotated(yRot / 16.0, 0.0, 1.0, 0.0);
  glRotated(zRot / 16.0, 0.0, 0.0, 1.0);

  glBindTexture(GL_TEXTURE_2D, texture[0]); // choose the texture to use.
  if(object)
    glCallList(object);
}

void GLWidget::resizeGL(int width, int height)
{
  // Reset The Current Viewport And Perspective Transformation
  glViewport(0, 0, width, height);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();

  gluPerspective(45.0f, (GLfloat)width/(GLfloat)height, 0.1f, 100.0f);
  glMatrixMode(GL_MODELVIEW);
}

void GLWidget::mousePressEvent(QMouseEvent *event)
{
  lastPos = event->pos();
}

void GLWidget::mouseMoveEvent(QMouseEvent *event)
{
  int dx = event->x() - lastPos.x();
  int dy = event->y() - lastPos.y();

  if(event->buttons() & Qt::LeftButton)
  {
    setXRotation(xRot + 8 * dy);
    setYRotation(yRot + 8 * dx);
  } 
  else if(event->buttons() & Qt::RightButton)
  {
    setXRotation(xRot + 8 * dy);
    setZRotation(zRot + 8 * dx);
  }

  lastPos = event->pos();
}

GLuint GLWidget::makeObject()
{
  float texW = float(this->frameWidth) / this->bufferWidth;
  float texH = float(this->frameHeight) / this->bufferHeight;
  float r = float(this->frameWidth) / this->frameHeight - 1.0;
  GLuint list = glGenLists(1);
  glNewList(list, GL_COMPILE);

  glBegin(GL_QUADS);
    glTexCoord2f(0.0f, texH);
    glVertex3f(-1.0-r, -1.0, 0.0f); // Bottom Left

    glTexCoord2f(texW, texH);
    glVertex3f( 1.0+r, -1.0, 0.0f); // Bottom Right

    glTexCoord2f(texW, 0.0f);
    glVertex3f( 1.0+r,  1.0, 0.0f); // Top Right

    glTexCoord2f(0.0f, 0.0f);
    glVertex3f(-1.0-r,  1.0, 0.0f); // Top Left
  glEnd();

  glEndList();
  return list;
}


void GLWidget::normalizeAngle(int *angle)
{
  while (*angle < 0)
    *angle += 360 * 16;
  while (*angle > 360 * 16)
    *angle -= 360 * 16;
}
