/**
 * YUV to RGB conversion
 */

#include "yuv2rgb.h"

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


#define CLIP_SIZE   811
#define CLIP_OFFSET 277

#define YMUL  298
#define RMUL  409
#define BMUL  516
#define G1MUL -100
#define G2MUL -208

static int           yuv2rgb_y[256];
static int           yuv2rgb_r[256];
static int           yuv2rgb_b[256];
static int           yuv2rgb_g1[256];
static int           yuv2rgb_g2[256];
static unsigned long yuv2rgb_clip[CLIP_SIZE];
static unsigned long yuv2rgb_clip8[CLIP_SIZE];
static unsigned long yuv2rgb_clip16[CLIP_SIZE];
static int           yuv2rgb_tables_initialized = 0;

#define RED(y, v)     yuv2rgb_clip16[CLIP_OFFSET + yuv2rgb_y[y] + yuv2rgb_r[v]]
#define GREEN(y,v,u)  yuv2rgb_clip8[CLIP_OFFSET + yuv2rgb_y[y] + yuv2rgb_g1[v] + yuv2rgb_g2[u]]
#define BLUE(y,u)     yuv2rgb_clip[CLIP_OFFSET + yuv2rgb_y[y] + yuv2rgb_b[u]]

static void init_yuv2rgb_tables(void)
{
  int i;

  for(i = 0; i < 256; i++)
    {
      yuv2rgb_y[i] = (YMUL * (i - 16) + 128) >> 8;
      yuv2rgb_r[i] = (RMUL * (i - 128)) >> 8;
      yuv2rgb_b[i] = (BMUL * (i - 128)) >> 8;
      yuv2rgb_g1[i] = (G1MUL * (i - 128)) >> 8;
      yuv2rgb_g2[i] = (G2MUL * (i - 128)) >> 8;
    }
  for(i = 0 ; i < CLIP_OFFSET; i++)
    {
      yuv2rgb_clip[i] = 0;
      yuv2rgb_clip8[i] = 0;
      yuv2rgb_clip16[i] = 0;
    }
  for( ; i < CLIP_OFFSET + 256; i++)
    {
      yuv2rgb_clip[i] = i - CLIP_OFFSET;
      yuv2rgb_clip8[i] = (i - CLIP_OFFSET) << 8;
      yuv2rgb_clip16[i] = (i - CLIP_OFFSET) << 16;
    }
  for( ; i < CLIP_SIZE; i++)
    {
      yuv2rgb_clip[i] = 255;
      yuv2rgb_clip8[i] = 255 << 8;
      yuv2rgb_clip16[i] = 255 << 16;
    }

  yuv2rgb_tables_initialized = 1;
}


/**
 * Faster function to convert YUV420 images to RGB32
 * RGB32: 0xFFRRGGBB
 * This function uses precalculated tables that are initialized
 * on the first run.
 * Make sure the src and dest buffers have enough room
 * dest should be width * height * 4 bytes in size
 * Based on the formulas found at http://en.wikipedia.org/wiki/YUV
 */
void YUV420_to_RGB32(int width, int height, char *src, char *dest)
{
  unsigned char *y, *u, *v;
  unsigned int *dst;
  int i;
  unsigned char *uu,*vv;
  unsigned int *d;
  int j;

  if(!yuv2rgb_tables_initialized)
    init_yuv2rgb_tables();

  dst = (unsigned int *)dest;
  y  = (unsigned char *)src;
  u  = y + width * height;
  v  = u + width * height / 4;

  for(i = 0; i < height; i++)
    {
      d = dst;
      uu = u;
      vv = v;
      for(j = 0; j < width >> 1; j++)
        {
          *(d++) = 0xff000000 |
            (RED(*y,*vv)>>16) |
            GREEN(*y,*vv,*uu) |
            (BLUE(*y,*uu)<<16);
          y++;
          *(d++) = 0xff000000 |
            (RED(*y,*vv)>>16) |
            GREEN(*y,*vv,*uu) |
            (BLUE(*y,*uu)<<16);
          y++;
          uu++;
          vv++;
        }
      if(i & 1)
        {
          u += width >> 1;
          v += width >> 1;
        }
      dst += width;
    }
}

#ifdef __cplusplus
}
#endif /* __cplusplus */
